%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% sample_bidirected_step:
%
% Update one row/column of a covariance matrix (including the sparsity
% pattern).
%
% INPUT:
%
% - N: sample size
% - RR: sufficient statistic (inner product of residuals)
% - delta: prior equivalent sample size for covariance matrix
% - U: prior inverse scale matrix
% - C: current value of covariance matrix
% - z_B: current indicators of sparsity
% - zps_B: current prior probability of sparsity
% - fixG: don't change edges marked with 1s
%
% OUTPUT:
%
% - updated values of C and z_B
%
% Initial version: Ricardo Silva, UCL - London, March 2011
% Current version: 26/10/2011
%
% * DISCLAIMER * DISCLAIMER * DISCLAIMER * DISCLAIMER * DISCLAIMER *
% * DISCLAIMER * DISCLAIMER * DISCLAIMER * DISCLAIMER * DISCLAIMER *
% * DISCLAIMER * DISCLAIMER * DISCLAIMER * DISCLAIMER * DISCLAIMER *
%
% This code is *NOT* meant to be efficient, but just to illustrate the main
% idea of the method. It could be much improved.

function [C z_B] = sample_bidirected_step(N, RR,...
                                          delta, U, C, z_B, zps_B, fixG)

num_y = size(RR, 2);

p = [0; 0];
MM = zeros(num_y, 1);

for i = 1:num_y
    
    not_i = [1:(i - 1) (i + 1):num_y];
    
    % Get initial Bartlett parameterization

    MM(not_i) = U(not_i, not_i) \ U(not_i, i); 
    pre_U_i   = MM(not_i)' * U(not_i, not_i) * MM(not_i);
    u_ii      = U(i, i) - U(i, not_i) * MM(not_i);    
    
    % Main loop
    
    for j = 1:(i - 1)
        
        if fixG(i, j)
           continue
        end

        pv = [1 - zps_B(i) * zps_B(j); zps_B(i) * zps_B(j)];
        %pv = [0.5 0.5];
                
        for z_value = [0 1]
            
            z_B(i, j) = z_value; z_B(j, i) = z_value; 
            
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            % Get basic spouse/non-spouse information

            sp  = z_B(i, :) == 1;             num_sp  = sum(sp);
            nsp = z_B(i, :) == 0; nsp(i) = 0; num_nsp = sum(nsp);
            A   = (C(nsp, nsp) \ C(nsp, sp))';
            
            U_ss = U(sp, sp);   
            U_ns = U(nsp, sp); 
            U_nn = U(nsp, nsp); 

            sub1  = U_ss  - A * U_ns;
            sub2  = U_ns' - A * U_nn;
            K_inv = sub1 - sub2 * A';
            K_chol_inv = chol(K_inv);
            K     = solve_chol(K_chol_inv, eye(num_sp));
            pre_m = [sub1 sub2] * [MM(sp) ; MM(nsp)];
            m     = solve_chol(K_chol_inv, pre_m);
               
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            % Get residual sufficient statistics
            
            ZZ = RR(sp, sp) ...
                   - A * RR(nsp, sp) - RR(sp, nsp) * A' ...
                   + A * RR(nsp, nsp) * A';
            YZ = RR(sp, i) - A * RR(nsp, i);
            YY = RR(i, i);

            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            % Get remainder prior for "g"
            
            g1 = (delta + num_y - 1 + num_nsp) / 2;
            g2 = (u_ii + pre_U_i - pre_m' * m) / 2;

            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            % Get "posterior sufficient statistics"
            
            T = K_inv + ZZ;
            q = YZ + pre_m;
            
            g1_prime = 0.5 * N + g1;
            g2_prime = 0.5 * (YY + m' * pre_m - q' * (T \ q)) + g2;
            if g2_prime < 0
              error('!')
            end
            
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            % Combine prior with likelihood to calculate probability

            p(z_value + 1) = log(pv(z_value + 1)) ...
                             - 0.5 * log(det(K)) - 0.5 * log(det(T)) ...
                             + g1 * log(g2) - gammaln(g1) ...
                             + gammaln(g1_prime) - g1_prime * log(g2_prime);
            
        end
        
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        % Perform sampling

        p = exp(p - max(p));
        p = p / (p(1) + p(2));
        
        if rand() < p(1)
           z_B(i, j) = 0; z_B(j, i) = 0;
        else
           z_B(i, j) = 1; z_B(j, i) = 1;
        end
        
    end
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % REFRESH STEP: refresh free coefficients 

    sp  = z_B(i, :) == 1;
    nsp = z_B(i, :) == 0; nsp(i) = 0;

    C(i, :) = update_g_invwishart_row(i, U + RR, delta + N, ...
                                      C, sp, nsp, not_i);
    C(:, i) = C(i, :);

end

% * DISCLAIMER * DISCLAIMER * DISCLAIMER * DISCLAIMER * DISCLAIMER *
% * DISCLAIMER * DISCLAIMER * DISCLAIMER * DISCLAIMER * DISCLAIMER *
% * DISCLAIMER * DISCLAIMER * DISCLAIMER * DISCLAIMER * DISCLAIMER *
%
% This code is *NOT* meant to be efficient, but just to illustrate the main
% idea of the method. It could be much improved.
