%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% sem_spikebd_sampler:
%
% Samples the "bi-directed structure" of a measurement model. That is, it
% does not change the structure of the linear coefficients, but the
% correlation structure among error terms in the measurement model is
% sampled.
%
% Input:
%
% - Y: a N-by-num_y matrix containing observations
% - G: a num_y-by-num_x binary matrix providing the directed structure
% - fix_W: a num_y-by-(num_x + 1) binary matrix indicating which loadings
%          (including the intercept at column num_x + 1) are fixed to a
%          given value
% - fix_W_v: the corresponding fixed values for each entry where 
%            fix_W(y, x) == 1
% - fix_GB: a num_y-by-num_y binary matrix indicating which bi-directed
%           edges are fixed to exist or not exist
% - fix_GB_g: a num_y-by-num_y binary matrix indicating if a bi-directed
%             edge exists or not if it is fixed
% - M: number of MCMC iterations
% - pw: prior precision for the Gaussian prior for loadings
% - u_B, U_B, u_X, U_X: hyperparameters for the prior of the error
%   covariance structure. Please see "simulate_latent_sem" for details
% - missing: a N-by-num_y binary matrix indicating which values are missing 
%            in the data
% - burn_miss: skip this much when returning the missing data estimator
%
% Output:
%
% - z_B: posterior samples for the binary indicators of the presence of 
%        bi-directed edges
% - w: posterior samples for the loadings
% - C_B: posterior samples for the error covariance matrix
% - C_X: posterior samples for the latent covariance matrix
% - miss_mean: posterior mean estimator for missing data (
%
% Initial version: Ricardo Silva, UCL - London, April 2011
% Current version: 1.0 (25/02/2011)

function [z_B w C_B C_X miss_mean] = ...
           sem_spikebd_sampler(Y, G, fix_W, fix_W_v, fix_GB, fix_GB_g, M, ...
                               pw, u_B, U_B, u_X, U_X, missing, burn_miss)

[N num_y] = size(Y);
num_x = size(G, 2);

if nargin < 13
  missing = zeros(size(Y));
  burn_miss = 0;
  num_missing = 0;
  miss_mean = [];
else
  num_missing = sum(sum(missing));   
  miss_mean = zeros(size(Y));
end

z = ones(num_y, num_x + 1);
w = zeros(num_y, num_x + 1, M); w(:, :, 1) = randn(num_y, num_x + 1);
fix_loadings = cell(num_y, 1);
for y = 1:num_y
  fix_z0 = ((fix_W(y, :) == 1) .* (fix_W_v(y, :) == 0)) == 1;
  z(y, fix_z0 == 1) = 0;
  z(y, G(y, :) == 0) = 0;
  fix_loadings{y} = fix_W(y, :) == 1;
  w(y, fix_loadings{y}, 1) = fix_W_v(y, fix_loadings{y});
  w(y, z(y, :, 1) == 0, 1) = 0;
end

z_B = zeros(num_y, num_y, M);
z_B(fix_GB == 1) = fix_GB_g(fix_GB == 1);

C_B = zeros(num_y, num_y, M);
for y = 1:num_y
  C_B(y, y, 1) = abs(randn()) +  0.5;
end
invC_B = inv(C_B(:, :, 1));

zps = zeros(num_y, M);
zps(:, 1) = 0.5;

X = [randn(N, num_x) ones(N, 1)];

if num_missing > 0
  Y_sim = randn(N, num_y) .* repmat(sqrt(diag(C_B(:, :, 1)))', N, 1) + X * w(:, :, 1)';
  Y(missing == 1) = Y_sim(missing == 1);
end

other_x = cell(num_x + 1, 1);
for x = 1:(num_x + 1)
  other_x{x} = [1:(x - 1) (x + 1):(num_x + 1)];
end

not_y = cell(num_y, 1);
for y = 1:num_y
  not_y{y} = [1:(y - 1) (y + 1):num_y];
end

C_X = zeros(num_x, num_x, M);
C_X(:, :, 1) = eye(num_x);
invC_X = inv(C_X(:, :, 1));

if burn_miss < 1
  burn_miss = 1;
end

num_z1 = sum(z(y, :, 1));

for i = 2:M
    
  % Initialize iteration
    
  if mod(i, 10) == 0
    fprintf('Iteration %d\n', i);
  end
    
  w(:, :, i)    = w(:, :, i - 1);
  zps(:, i)    = zps(:, i - 1);
  z_B(:, :, i) = z_B(:, :, i - 1);
  C_B(:, :, i) = C_B(:, :, i - 1);
  C_X(:, :, i) = C_X(:, :, i - 1);

  % Sample latent variables
    
  inv_covX = w(:, 1:num_x, i)' * invC_B * w(:, 1:num_x, i) + invC_X;
  covX   = inv(inv_covX);
  coeffX = inv_covX \ w(:, 1:num_x, i)' * invC_B;
  meanX  = (Y - repmat(w(:, end, i)', N, 1)) * coeffX';
  X(:, 1:num_x) = mvnrnd(meanX, covX);

  XX = X' * X;
  XY = X' * Y;

  % Sample latent covariance matrix
    
  pos_u = u_X + N;
  pos_U = U_X + XX(1:num_x, 1:num_x);    
  Z = mvnrnd(zeros(pos_u - num_x + 1, num_x), inv(pos_U))';
  invC_X = Z * Z';
  C_X(:, :, i) = inv(invC_X);

  % Sample measurement model
    
  for y = 1:num_y
        
    z_sel = find(z(y, :) == 1);
    inv_newCov = XX(z_sel, z_sel) * invC_B(y, y) + pw * eye(num_z1);
    newCov = inv(inv_newCov);
    newM   = inv_newCov \ (invC_B(y, y) * XY(z_sel, y) + ...
             (XY(z_sel, not_y{y}) - XX(z_sel, :) * w(not_y{y}, :, i)') * ...
             invC_B(not_y{y}, y)) ;
    if isempty(fix_loadings{y})
      w(y, z_sel, i) = randnorm(1, newM, [], newCov);
    else                       
      z_fixed  = ((fix_W(y, z_sel) == 1) .* z(y, z_sel)) == 1;
      z_free   = ((fix_W(y, z_sel) == 0) .* z(y, z_sel)) == 1;
      coeff    = newCov(z_fixed, z_fixed) \ newCov(z_fixed, z_free);
      condvar  = newCov(z_free, z_free) - newCov(z_free, z_fixed) * coeff;
      condmean = newM(z_free) + coeff' * (w(y, z_fixed, i)' - newM(z_fixed)); 
      w(y, z_sel(z_free), i) = mvnrnd(condmean, condvar)';
    end
      
  end

  R  = Y - X * w(:, :, i)';
  RR = R' * R;
  [C_B(:, :, i) z_B(:, :, i)] = ...
       sample_bidirected_step(N, RR, u_B, U_B, ...
             C_B(:, :, i), z_B(:, :, i), zps(:, i), fix_GB);
  invC_B = inv(C_B(:, :, i));

  if num_missing > 0 && i > burn_miss
    miss_mean = miss_mean + Y;
  end
    
end

miss_mean = miss_mean / (M - burn_miss);

